class Tracer:
    
    def __init__(self, ):
        
        pass
    
    def _trace():
        
        pass
    
    def trace():
        
        pass
    
    def plot():
        
        pass
    
    def writeto():
        
        pass